unit GameTypes;

interface

uses
  SysUtils,
  Classes,
  Windows,
  Lapage,
  BaseFile,
  GameConstr;

type
  TGameArea = class;
  TGirl = class;

  TGame = class
  private
    FArea: TGameArea;
    FTilesSpt: TLpgList;
    FArrowsSpt: TLpgList;
    FUnits: TList;
    FGirlsSpt: TLpgList;
    FBonusStack: TList;
    FBonusStackSptOffset: Integer; //   
  public
    constructor Create;
    destructor Destroy; override;

    procedure LoadLevel(AlevelFileName: string);

    property Area: TGameArea read FArea;
    procedure Draw(Buff: TLpgRowsBuffer);

    function Action: Boolean;

    procedure Click(X, Y: Integer);
  end;

  TGameCell = class
  private
    FArea: TGameArea;
    FOrientation: Byte;
    FOrientationExists: Boolean;
    FTileIndex: Integer;
  public
    constructor Create(AArea: TGameArea);
    destructor Destroy; override;
  end;

  TGameArea = class
  private
    FGame: TGame;
    FWidth: Word;
    FHeight: Word;
    FCells: array of array of TGameCell;
    FOffsetX: Integer;
    FOffsetY: Integer;
  public
    constructor Create(AGame: TGame);
    destructor Destroy; override;

    function isValidCell(cx, cy : Integer) : Boolean;

    procedure SetSize(ANewWidth, ANewHeight: Word);
    property Width: Word read FWidth;
    property Height: Word read FHeight;
    property OffsetX: Integer read FOffsetX;
    property OffsetY: Integer read FOffsetY;

    procedure Draw(ABuff: TLpgRowsBuffer);
  end;

  TGameUnit = class
  private
    FGame: TGame;
    FArea: TGameArea;

    FX: Double;
    FY: Double;
    FCellLastX: Integer;
    FCellLastY: Integer;
    FCellX: Integer;
    FCellY: Integer;
    procedure SetCellXY(CX, CY : Integer); inline;
    procedure SetX(const Value: Double);
    procedure SetY(const Value: Double);
    function GetSpriteIndex: Integer; virtual;
  protected
    procedure ChangeCell(Cx, Cy : Integer);virtual;
  public
    constructor Create(AGame: TGame);
    destructor Destroy; override;

    procedure Draw(ABuff: TLpgRowsBuffer; X, Y: Integer);

    property SpriteIndex: Integer read GetSpriteIndex;

    property X: Double read FX write SetX;
    property Y: Double read FY write SetY;

    property CellX: Integer read FCellX;
    property CellY: Integer read FCellY;

    function Action: Boolean; virtual;
  end;

  TGirl = class(TGameUnit)
  private
    FAnimationOffset : Byte;
    FOrientation: Integer;
    FSpeed: Double;
    function GetSpriteIndex: Integer; override;
    procedure SetOrientation(const Value: Integer);
    procedure SetSpeed(const Value: Double);
  protected
    procedure ChangeCell(Cx, Cy : Integer); override;
  public
    constructor Create(AGame: TGame);
    destructor Destroy; override;

    property Orientation: Integer read FOrientation write SetOrientation;
    property Speed: Double read FSpeed write SetSpeed;
    function Action: Boolean; override;
  end;

implementation

{ TGame }

function TGame.Action: Boolean;
var
  i: Integer;
begin
  for i := FUnits.Count - 1 downto 0 do
  begin
    if not TGameUnit(FUnits[i]).Action then
      TGameUnit(FUnits[i]).Free;
  end;

  //  
  FBonusStackSptOffset := FBonusStackSptOffset div 2;
  while FBonusStack.Count < 3 do
  begin
    FBonusStack.Add(TObject(Random(4)));
    inc(FBonusStackSptOffset, Arrow_Height);
  end;

  Result := true;
end;

procedure TGame.Click(X, Y: Integer);
var
  cx, cy: Integer;
begin
  cx := (X - Area.OffsetX) div Cell_Width;
  cy := (Y - Area.OffsetY) div Cell_Height;
  if FArea.isValidCell(cx, cy) then
  with FArea.FCells[cx, cy] do
  begin
    if FTileIndex <> 0 then
      exit;
    if FOrientationExists then
      exit;
    FOrientationExists := true;
    FOrientation := Integer(FBonusStack[0]);
    FBonusStack.Delete(0);
  end;
end;

constructor TGame.Create;
begin
  inherited;
  FArea := TGameArea.Create(self);
  FTilesSpt := LpgCreateCropList('data\tiles.png', $FF000000, Tile_Width,
    Tile_Height, LAPAGE_OBJECT_ROWBUFFER);
  FArrowsSpt := LpgCreateCropList('data\Arrows.png', $FF000000, Arrow_Width,
    Arrow_Height, LAPAGE_OBJECT_ROWBUFFER);

  FGirlsSpt := LpgCreateCropList('data\BLONDY.png', $FF000000, Girl_Spt_Width,
    Girl_Spt_Height, LAPAGE_OBJECT_ROWBUFFER);

  FUnits := TList.Create;
  FBonusStack := TList.Create;
end;

destructor TGame.Destroy;
var
  i: Integer;
begin
  if Assigned(FArea) then
    FArea.Free;
  if Assigned(FUnits) then
  begin
    for i := 0 to FUnits.Count - 1 do
    begin
      TObject(FUnits[0]).Free;
      // FUnits.Delete(0); -  
    end;
    FUnits.Free;
  end;

  if Assigned(FBonusStack) then
    FBonusStack.Free;

  FTilesSpt.Release;
  FArrowsSpt.Release;
  FGirlsSpt.Release;
  inherited;
end;

procedure TGame.Draw(Buff: TLpgRowsBuffer);
var
  i: Integer;
begin
  Area.Draw(Buff);
  //  
  for i := 0 to FBonusStack.Count - 1 do
  begin
    FArrowsSpt.Draw(Buff, 0, Arrow_Height * i + FBonusStackSptOffset, Integer
        (FBonusStack[i]))
  end;
end;

procedure TGame.LoadLevel(AlevelFileName: string);
var
  Base: TBaseFile;
begin
  Base := TBaseFile.Create;
  Base.LoadFromFile(AlevelFileName);

  Area.SetSize(Base.IntValue('Level.Width'), Base.IntValue('Level.Height'));

  Base.Free;
end;

{ TGameArea }

constructor TGameArea.Create(AGame: TGame);
begin
  inherited Create;
  FGame := AGame;
end;

destructor TGameArea.Destroy;
begin
  SetSize(0, 0);
  inherited;
end;

procedure TGameArea.Draw(ABuff: TLpgRowsBuffer);
var
  i, j, uIndex: Integer;
  g: TGameUnit;
  SptIndex: Integer;
  CurX, CurY: Integer;
begin
  CurX := (lpgMouseX - OffsetX) div Cell_Width;
  CurY := (lpgMouseY - OffsetY) div Cell_Width;

  for j := 0 to Height - 1 do
    for i := 0 to Width - 1 do
    begin
      //   
      SptIndex := FCells[i, j].FTileIndex;
      if (SptIndex = 0) and (i = CurX) and (j = CurY) and
        (not FCells[CurX, CurY].FOrientationExists) then
        SptIndex := 1;

      FGame.FTilesSpt.Draw(ABuff, i * Cell_Width + OffsetX,
        j * Cell_Height + OffsetY + Tile_Offset_Top, SptIndex);

      if FCells[i, j].FOrientationExists then
        FGame.FArrowsSpt.Draw(ABuff, i * Cell_Width + OffsetX + 1,
          j * Cell_Height + OffsetY + 1, FCells[i, j].FOrientation);
      if SptIndex = 1 then
        FGame.FArrowsSpt.Draw(ABuff, i * Cell_Width + OffsetX + 1,
          j * Cell_Height + OffsetY + 1, integer(FGame.FBonusStack[0]));
      //  ,    
      for uIndex := 0 to FGame.FUnits.Count - 1 do
      begin
        g := FGame.FUnits[uIndex];
        if g.CellY = j then
          g.Draw(ABuff, Round(g.X + OffsetX), Round(g.Y + OffsetY));
      end;
    end;
end;

function TGameArea.isValidCell(cx, cy: Integer): Boolean;
begin
  Result := (cx >= 0)and(cy >= 0)and(cx < Width)and(cy < Height);
end;

procedure TGameArea.SetSize(ANewWidth, ANewHeight: Word);
var
  i, j: Integer;
begin
  for i := 0 to Width - 1 do
    for j := 0 to Height - 1 do
    begin
      FCells[i, j].Free;
    end;

  FWidth := ANewWidth;
  FHeight := ANewHeight;

  if FWidth >= (Screen_Width div Cell_Width) then
    FWidth := (Screen_Width div Cell_Height) - 1;

  if FHeight >= (Screen_Height div Cell_Height) then
    FHeight := (Screen_Height div Cell_Height) - 1;

  SetLength(FCells, Width, Height);

  FOffsetX := (Screen_Width - Width * Cell_Width) div 2;
  FOffsetY := (Screen_Height - Height * Cell_Height) div 2;

  for i := 0 to Width - 1 do
    for j := 0 to Height - 1 do
    begin
      FCells[i, j] := TGameCell.Create(self);
    end;
end;

{ TGameCell }

constructor TGameCell.Create(AArea: TGameArea);
begin
  inherited Create;
  FArea := AArea;
  if random(6) = 2 then
    FTileIndex := Random(2)+2;
  if FTileIndex = 1 then
    FTileIndex := 0;
end;

destructor TGameCell.Destroy;
begin

  inherited;
end;

{ TGameUnit }

function TGameUnit.Action: Boolean;
begin
  Result := true;
end;

procedure TGameUnit.ChangeCell(Cx, Cy: Integer);
begin

end;

constructor TGameUnit.Create(AGame: TGame);
begin
  inherited Create;
  FGame := AGame;
  FArea := FGame.FArea;

  FGame.FUnits.Add(self);
end;

destructor TGameUnit.Destroy;
var
  index: Integer;
begin
  index := FGame.FUnits.IndexOf(self);
  if index <> -1 then
    FGame.FUnits.Delete(index);
  inherited;
end;

procedure TGameUnit.Draw(ABuff: TLpgRowsBuffer; X, Y: Integer);
begin
  FGame.FGirlsSpt.Draw(ABuff, X - 6, Y - 15, SpriteIndex);
end;

function TGameUnit.GetSpriteIndex: Integer;
begin
  Result := 0;
end;

procedure TGameUnit.SetCellXY(CX, CY: Integer);
begin
  if (Cx = FCellLastX)and(Cy = FCellLastY) then exit;
  FCellLastX := CX;
  FCellLastY := CY;
  ChangeCell(CX,CY);
end;

procedure TGameUnit.SetX(const Value: Double);
begin
  FX := Value;
  FCellX := Trunc(FX) div Cell_Width;
  SetCellXY(FCellX, FCellY);
end;

procedure TGameUnit.SetY(const Value: Double);
begin
  FY := Value;
  FCellY := Trunc(FY) div Cell_Height;
  SetCellXY(FCellX, FCellY);
end;

{ TGirl }

function TGirl.Action: Boolean;
begin
  if FArea.isValidCell(CellX, CellY) then
  begin
    if FArea.FCells[CellX, CellY].FOrientationExists then
    begin
      FOrientation := FArea.FCells[CellX, CellY].FOrientation;
      FArea.FCells[CellX, CellY].FOrientationExists := False;
    end;
  end
    else
  begin
    // !
    Exit(false);
  end;

  case Orientation of
    Orientation_Up:
      Y := Y - Speed;
    Orientation_Right:
      X := X + Speed;
    Orientation_Down:
      Y := Y + Speed;
    Orientation_Left:
      X := X - Speed;
  end;
  if Orientation in [Orientation_Up, Orientation_Down] then
    X := CellX * Cell_Width + (Cell_Width div 2)
  else
    Y := CellY * Cell_Height + (Cell_Height div 2);

  Result := true;
end;

procedure TGirl.ChangeCell(Cx, Cy: Integer);
begin
  inherited;

end;

constructor TGirl.Create(AGame: TGame);
begin
  inherited Create(AGame);
  FAnimationOffset := Random(256);
  X := 10;
  Y := 10;
  Speed := 0.8;
  Orientation := 1;
end;

destructor TGirl.Destroy;
begin

  inherited;
end;

function TGirl.GetSpriteIndex: Integer;
begin
  Result := (((GetTickCount div 200) + FAnimationOffset) mod 4) + Orientation * 4;
end;

procedure TGirl.SetOrientation(const Value: Integer);
begin
  FOrientation := Value;
end;

procedure TGirl.SetSpeed(const Value: Double);
begin
  FSpeed := Value;
end;

end.
