#include "OpenGL.h"

#include <System/Log/Log.h>

namespace Graphics
{
	Types::Property<const char *> OpenGL::Subsystem("Graphics/OpenGL", true);
};

Graphics::OpenGL::OpenGL()
{
	Wnd = nullptr;
	glDC = 0;
	glRC = 0;
}

Graphics::OpenGL::~OpenGL()
{
	if (glRC)
	{
		wglMakeCurrent(glDC,0);
		wglDeleteContext(glRC);
		glRC = 0;
	}
	if (glDC)
	{
		__try
		{
			if (Wnd)
				ReleaseDC(Wnd->GetHandle(), glDC);
		}
		__except(EXCEPTION_EXECUTE_HANDLER)
		{
		}
		glDC = 0;
	}
}

HDC Graphics::OpenGL::GetDC()
{
	return glDC;
}

HGLRC Graphics::OpenGL::GetRC()
{
	return glRC;
}

void Graphics::OpenGL::DumpInfo()
{
	System::Log::Message(Subsystem, "OpenGL vendor: %s", glGetString(GL_VENDOR));
	System::Log::Message(Subsystem, "OpenGL renderer: %s", glGetString(GL_RENDERER));
	System::Log::Message(Subsystem, "OpenGL version: %s", glGetString(GL_VERSION));
	System::Log::Message(Subsystem, "OpenGL shading language version: %s", glGetString(GL_SHADING_LANGUAGE_VERSION));
	System::Log::Message(Subsystem, "OpenGL extensions: %s", glGetString(GL_EXTENSIONS));
	System::Log::Message(Subsystem, "GLew version: %s", glewGetString(GLEW_VERSION));
}

bool Graphics::OpenGL::Init(Graphics::Window & Wnd)
{
	this->Wnd = &Wnd;

	if (Wnd.GetHandle() == 0)
	{
		System::Log::CriticalError(Subsystem, "Cannot initialize OpenGL on uninitialized window");
		return false;
	}

	glDC = Wnd.GetDC();

	PIXELFORMATDESCRIPTOR pfd;
	memset(&pfd, 0, sizeof(pfd));

	pfd.nSize = sizeof(pfd);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pfd.iLayerType = PFD_MAIN_PLANE;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 24;
	pfd.cDepthBits = 16;

	int PixelFormat = ChoosePixelFormat(glDC, &pfd);
	if (!PixelFormat)
	{
		System::Log::CriticalError(Subsystem, "Cannot choose pixel format");
		return false;
	}

	if (!SetPixelFormat(glDC, PixelFormat, &pfd))
	{
		System::Log::CriticalError(Subsystem, "Cannot set pixel format");
		return false;
	}

	glRC = wglCreateContext(glDC);
	if (!glRC)
	{
		System::Log::CriticalError(Subsystem, "Cannot create rendering context");
		return false;
	}

	wglMakeCurrent(glDC, glRC);

	SetDefaultParameters();

	if (glewInit() != GLEW_OK)
	{
		System::Log::CriticalError(Subsystem, "Cannot initialize GLew");
		return false;
	} else {
		System::Log::Message(Subsystem, "GLew was initialized");
	}

	return true;
}

void Graphics::OpenGL::SetDefaultParameters()
{
	glClearColor(0.0, 0.0, 0.0, 0.0);

	glEnable(GL_COLOR_MATERIAL);

	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POLYGON_SMOOTH);
	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	Set2D();
}

void Graphics::OpenGL::Clear()
{
	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
}

void Graphics::OpenGL::SwapBuffers()
{
	::SwapBuffers(glDC);
}

void Graphics::OpenGL::Set2D()
{
	if (!Wnd)
		return;

	int w, h;
	__try
	{
		w = Wnd->GetClientWidth();
		h = Wnd->GetClientHeight();
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		return;
	}

	glDisable(GL_DEPTH_TEST);

	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, w, h, 0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}