#include "Shader.h"

#include <System/Log/Log.h>

namespace Graphics
{
	Types::Property<const char *> Shader::Subsystem("Graphics/Shader", true);
};

Graphics::Shader::Shader()
{
	Handle = 0;
	Initialized = false;
}

Graphics::Shader::~Shader()
{
	Delete();
}

bool Graphics::Shader::LoadFromString(GLenum ShaderType, const char * Source)
{
	__try
	{
		Handle = glCreateShaderObjectARB(ShaderType);
		Initialized = true;
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		System::Log::Error(Subsystem, "Cannot create shader object (type %x)", ShaderType);
		return false;
	}

	GLint len = strlen(Source);

	glShaderSourceARB(Handle, 1, &Source, &len);
	glCompileShaderARB(Handle);

	if (glGetError() != GL_NO_ERROR)
	{
		System::Log::Error(Subsystem, "Cannot compile shader (type %x), error %d: %s", ShaderType, glGetError(), gluErrorString(glGetError()));

		Delete();
		return false;
	}

	return true;
}

bool Graphics::Shader::LoadFromFile(GLenum ShaderType, const char * Filename)
{
	FILE * file;
	if (fopen_s(&file, Filename, "rb"))
	{
		System::Log::Error(Subsystem, "Cannot load shader \"%s\" (type %x): file not found", Filename, ShaderType);
		return false;
	}

	fseek(file, 0, SEEK_END);
	int size = ftell(file);

	if (size < 1)
	{
		fclose(file);
		System::Log::Error(Subsystem, "Cannot load shader \"%s\" (type %x): file is empty", Filename, ShaderType);
		return false;
	}

	char * buf = (char *)malloc(size);
	if (!buf)
	{
		fclose(file);
		System::Log::Error(Subsystem, "Cannot load shader \"%s\" (type %x): memory allocation failed", Filename, ShaderType);
		return false;
	}

	memset(buf, 0, size);

	fseek(file, 0, SEEK_SET);

	if (int read = fread_s(buf, size, 1, size, file) != size)
	{
		free(buf);
		fclose(file);
		System::Log::Error(Subsystem, "Cannot load shader \"%s\" (type %x): read error (read %d, size %d)", Filename, ShaderType, read, size);
		return false;
	}

	fclose(file);

	__try
	{
		Handle = glCreateShaderObjectARB(ShaderType);
		Initialized = true;
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		free(buf);
		System::Log::Error(Subsystem, "Cannot load shader \"%s\" (type %x): cannot create shader object", Filename, ShaderType);
		return false;
	}

	glShaderSourceARB(Handle, 1, (const GLcharARB **)&buf, &size);

	free(buf);

	glCompileShaderARB(Handle);

	if (glGetError() != GL_NO_ERROR)
	{
		System::Log::Error(Subsystem, "Cannot compile shader \"%s\" (type %x), error %d: %s", Filename, ShaderType, glGetError(), gluErrorString(glGetError()));

		Delete();
		return false;
	}

	System::Log::Message(Subsystem, "Shader \"%s\" (type %x) was loaded", Filename, ShaderType);

	return true;
}

bool Graphics::Shader::IsLoaded()
{
	return Initialized;
}

GLhandleARB Graphics::Shader::GetHandle()
{
	return Handle;
}

void Graphics::Shader::Delete()
{
	if (Initialized)
	{
		glDeleteShader(Handle);
		Handle = 0;
		Initialized = false;
	}
}

bool Graphics::VertexShader::LoadFromString(const char * Source)
{
	return ((Shader *)this)->LoadFromString(GL_VERTEX_SHADER_ARB, Source);
}

bool Graphics::VertexShader::LoadFromFile(const char * Source)
{
	return ((Shader *)this)->LoadFromFile(GL_VERTEX_SHADER_ARB, Source);
}

bool Graphics::FragmentShader::LoadFromString(const char * Source)
{
	return ((Shader *)this)->LoadFromString(GL_FRAGMENT_SHADER_ARB, Source);
}

bool Graphics::FragmentShader::LoadFromFile(const char * Source)
{
	return ((Shader *)this)->LoadFromFile(GL_FRAGMENT_SHADER_ARB, Source);
}