#include "ShaderProgram.h"

#include <System/Log/Log.h>

namespace Graphics
{
	Types::Property<const char *> ShaderProgram::Subsystem("Graphics/ShaderProgram", true);
};

Graphics::ShaderProgram::ShaderProgram()
{
	Initialized = false;
	Handle = 0;
}

Graphics::ShaderProgram::~ShaderProgram()
{
	Delete();
}

GLhandleARB Graphics::ShaderProgram::GetHandle()
{
	if (!Initialized)
	{
		Handle = glCreateProgramObjectARB();
		Initialized = true;
	}

	return Handle;
}

bool Graphics::ShaderProgram::Attach(Graphics::Shader & Shader)
{
	if (!Shader.IsLoaded())
		return false;

	GLhandleARB program = GetHandle();

	glAttachObjectARB(program, Shader.GetHandle());

	if (glGetError() != GL_NO_ERROR)
	{
		System::Log::Error(Subsystem, "Cannot attach shader to shader program, error %d: %s", glGetError(), gluErrorString(glGetError()));

		return false;
	}

	return true;
}

bool Graphics::ShaderProgram::Link()
{
	GLhandleARB program = GetHandle();

	glLinkProgramARB(program);

	if (glGetError() != GL_NO_ERROR)
	{
		System::Log::Error(Subsystem, "Cannot link shader program, error %d: %s", glGetError(), gluErrorString(glGetError()));

		return false;
	}

	return true;
}

void Graphics::ShaderProgram::Delete()
{
	if (Initialized)
	{
		glDeleteProgramsARB(1, &Handle);
		Handle = 0;
		Initialized = false;
	}
}

void Graphics::ShaderProgram::Bind()
{
	GLhandleARB program = GetHandle();

	glUseProgramObjectARB(program);
}

void Graphics::ShaderProgram::Unbind()
{
	glUseProgramObjectARB(0);
}

bool Graphics::ShaderProgram::GetUniformFloat(const char * Name, float * Buf)
{
	GLhandleARB program = GetHandle();

	int loc = glGetUniformLocationARB(program, Name);
	if (loc < 0)
	{
		System::Log::Warning(Subsystem, "Cannot get uniform \"%s\" on program %d", Name, program);
		return false;
	}

	glGetUniformfvARB(program, loc, Buf);

	if (GLenum Error = glGetError() != GL_NO_ERROR)
	{
		System::Log::Warning(Subsystem, "Cannot get uniform value \"%s\" on program %d, error %d: %s", Name, program, Error, gluErrorString(Error));

		return false;
	}

	return true;
}

bool Graphics::ShaderProgram::GetUniformInt(const char * Name, int * Buf)
{
	GLhandleARB program = GetHandle();

	int loc = glGetUniformLocationARB(program, Name);
	if (loc < 0)
	{
		System::Log::Warning(Subsystem, "Cannot get uniform \"%s\" on program %d", Name, program);
		return false;
	}

	glGetUniformivARB(program, loc, Buf);

	if (GLenum Error = glGetError() != GL_NO_ERROR)
	{
		System::Log::Warning(Subsystem, "Cannot get uniform value \"%s\" on program %d, error %d: %s", Name, program, Error, gluErrorString(Error));

		return false;
	}

	return true;
}

bool Graphics::ShaderProgram::SetUniformFloat(const char * Name, int Count, float * Value)
{
	if (Count < 1 || Count > 4)
		return false;

	GLhandleARB program = GetHandle();

	int loc = glGetUniformLocationARB(program, Name);
	if (loc < 0)
	{
		System::Log::Warning(Subsystem, "Cannot get uniform \"%s\" on program %d", Name, program);
		return false;
	}

	switch(Count)
	{
	case 1:
		glUniform1fvARB(loc, 1, Value);
		break;
	case 2:
		glUniform2fvARB(loc, 1, Value);
		break;
	case 3:
		glUniform3fvARB(loc, 1, Value);
		break;
	case 4:
		glUniform4fvARB(loc, 1, Value);
		break;
	}

	if (GLenum Error = glGetError() != GL_NO_ERROR)
	{
		System::Log::Warning(Subsystem, "Cannot set uniform value \"%s\" on program %d, error %d: %s", Name, program, Error, gluErrorString(Error));

		return false;
	}

	return true;
}

bool Graphics::ShaderProgram::SetUniformInt(const char * Name, int Count, int * Value)
{
	if (Count < 1 || Count > 4)
		return false;

	GLhandleARB program = GetHandle();

	int loc = glGetUniformLocationARB(program, Name);
	if (loc < 0)
	{
		System::Log::Warning(Subsystem, "Cannot get uniform \"%s\" on program %d", Name, program);
		return false;
	}

	switch(Count)
	{
	case 1:
		glUniform1ivARB(loc, 1, Value);
		break;
	case 2:
		glUniform2ivARB(loc, 1, Value);
		break;
	case 3:
		glUniform3ivARB(loc, 1, Value);
		break;
	case 4:
		glUniform4ivARB(loc, 1, Value);
		break;
	}

	if (GLenum Error = glGetError() != GL_NO_ERROR)
	{
		System::Log::Warning(Subsystem, "Cannot set uniform value \"%s\" on program %d, error %d: %s", Name, program, Error, gluErrorString(Error));

		return false;
	}

	return true;
}

GLint Graphics::ShaderProgram::GetUniformLocation(const char * Name)
{
	GLhandleARB program = GetHandle();

	int loc = glGetUniformLocationARB(program, Name);

	return loc;
}
