#include "Window.h"

#include <Consts/Consts.h>
#include <System/Log/Log.h>
#include <System/Application/Application.h>

namespace Graphics
{
	Types::Property<const char *> Window::Subsystem("Graphics/Window", true);
	unsigned long Window::Instances = 0;

	System::Threads::CriticalSection Window::CritSection;

	ATOM Window::WndClass = 0;
};

Graphics::Window::Window()
{
	Handle = 0;

	CritSection.Enter();
	Instances++;
	CritSection.Leave();
}

Graphics::Window::~Window()
{
	if (Handle)
	{
		DestroyWindow(Handle);
		Handle = 0;
	}

	CritSection.Enter();

	Instances--;

	if (WndClass && Instances == 0)
	{
		UnregisterClass(Consts::EngineName, System::Application::Handle.Get());
		WndClass = 0;
	}

	CritSection.Leave();
}

bool Graphics::Window::Init(int Width, int Height, bool HasTitleBar, const char * Title, WNDPROC WndProc)
{
	if (WndClass == 0)
	{
		WNDCLASSEX wce;
		memset(&wce, 0, sizeof(wce));

		wce.cbSize			= sizeof(wce);
		wce.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
		wce.lpfnWndProc		= WndProc;
		wce.hInstance		= System::Application::Handle.Get();
		wce.hIcon			= LoadIcon(0, IDI_APPLICATION);
		wce.hCursor			= LoadCursor(0, IDC_ARROW);
		wce.hbrBackground	= (HBRUSH)GetStockObject(BLACK_BRUSH);
		wce.lpszClassName	= Consts::EngineName;

		CritSection.Enter();

		WndClass = RegisterClassEx(&wce);

		CritSection.Leave();

		if (!WndClass)
		{
			System::Log::CriticalError(Subsystem, "Cannot register window class");
			return false;
		}
	}

	Handle = CreateWindowEx(0, 
		Consts::EngineName,
		Title == nullptr ? Consts::EngineName : Title,
		HasTitleBar ? WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_VISIBLE : WS_POPUP | WS_VISIBLE,
		0, 0, Width, Height,
		0, 0, System::Application::Handle.Get(), nullptr);

	if (!Handle)
	{
		System::Log::CriticalError(Subsystem, "Cannot create window");
			return false;
	}

	ShowWindow(Handle, SW_SHOW);
	UpdateWindow(Handle);

	return true;
}

HWND Graphics::Window::GetHandle()
{
	if (Handle && !IsWindow(Handle)) Handle = 0;
	return Handle;
}

HDC Graphics::Window::GetDC()
{
	HWND Wnd = GetHandle();
	return Wnd == 0 ? 0 : ::GetDC(Wnd);
}

int Graphics::Window::GetWidth()
{
	RECT r;
	HWND h = GetHandle();
	if (!h) return 0;
	GetWindowRect(GetHandle(), &r);
	return r.right - r.left;
}

int Graphics::Window::GetHeight()
{
	RECT r;
	HWND h = GetHandle();
	if (!h) return 0;
	GetWindowRect(GetHandle(), &r);
	return r.bottom - r.top;
}

int Graphics::Window::GetClientWidth()
{
	RECT r;
	HWND h = GetHandle();
	if (!h) return 0;
	GetClientRect(GetHandle(), &r);
	return r.right - r.left;
}

int Graphics::Window::GetClientHeight()
{
	RECT r;
	HWND h = GetHandle();
	if (!h) return 0;
	GetClientRect(GetHandle(), &r);
	return r.bottom - r.top;
}