#include "Application.h"

#include <algorithm>

#include <System/Log/Log.h>

namespace System
{
	Types::Property<const char *> Application::Subsystem("System/Application", true);
	Types::Property<bool> Application::Running = false;

	bool Application::Initialized = false;
	bool Application::Finalized = false;

	System::Threads::CriticalSection Application::CritSection;

	std::deque<Application::Event *> Application::OnInitialize;
	std::deque<Application::Event *> Application::OnFinalize;

	Application::Event & Application::RunProc = Application::DefaultRunProc;

	Types::Property<HMODULE> Application::Handle(GetModuleHandle(0), true);
};

void AtExit_ApplicationTerminate()
{
	System::Application::Terminate();
}

void System::Application::CallEventList(std::deque<Event *> EventList)
{
	CritSection.Enter();
	std::for_each(EventList.begin(), EventList.end(), [&](Application::Event * Ev)
		{
			int code;
			__try
			{
				if (Ev != nullptr)
				{
					(*Ev)();
				}
			}
			// it's stupid, but working
			__except(code = _exception_code() || true)
			{
				System::Log::CriticalError(Subsystem, "Unhandled exception in event handler (code %d = 0x%x)", code, code);
			}
		}
	);
	CritSection.Leave();
}

void System::Application::DefaultRunProc()
{
	MSG Msg;
	while (Running)
	{
		if (PeekMessage(&Msg, 0, 0, 0, PM_REMOVE))
		{
			TranslateMessage(&Msg);
			DispatchMessage(&Msg);
		}
		Sleep(0);
	}
}

void System::Application::Run(Event & RunProc)
{
	static bool AtExitHandlerSet = false;
	if (!AtExitHandlerSet)
	{
		atexit(AtExit_ApplicationTerminate);
		AtExitHandlerSet = true;
	}

	if (Running)
	{
		System::Log::Warning(Subsystem, "Trying to run application more than 1 times at same time");
		return;
	}

	if (!Initialized)
	{
		CallEventList(OnInitialize);
		Initialized = true;
		Finalized = false;

		System::Log::Message(Subsystem, "Application has been initialized (%d initializators)", OnInitialize.size());
	}

	Running = true;
	
	RunProc();

	if (!Finalized)
	{
		CallEventList(OnFinalize);
		Finalized = true;
		Initialized = false;

		System::Log::Message(Subsystem, "Application has been finalized (%d finalizators)", OnFinalize.size());
	}

	exit(0);
}

void System::Application::Run()
{
	System::Application::Run(RunProc);
}

void System::Application::Terminate()
{
	if (Running)
		Running = false;
	else
	{
		System::Log::Warning(Subsystem, "Trying to terminate application, but application is not running");
		exit(0);
	}
}

void System::Application::AddInitializator(Event & Initializator)
{
	OnInitialize.push_back(&Initializator);
}

void System::Application::AddFinalizator(Event & Finalizator)
{
	OnFinalize.push_back(&Finalizator);
}

bool System::Application::IsRunning()
{
	return Running.Get();
}