#include "Log.h"

namespace System
{
	Types::Property<const char *> Log::Subsystem("System/Log", true);

#ifdef _DEBUG
	unsigned int Log::Level = Log::MessageType::Message | Log::MessageType::Warning | Log::MessageType::Error | Log::MessageType::CriticalError;
	bool Log::StopOnError = false;
#else
	unsigned int Log::Level = Log::MessageType::Message | Log::MessageType::Warning | Log::MessageType::Error | Log::MessageType::CriticalError;
	bool Log::StopOnError = false;
#endif
	FILE * Log::LogFile = nullptr;

	char Log::Buffer[16 * 1024] = {0};
	System::Threads::CriticalSection Log::CritSection;
};

#ifndef LOG_USE_COLORS

// Non-colored log messages
void System::Log::PrintMessage(const char * Subsystem, const char Prefix, const char * Text, va_list Args)
{
	CritSection.Enter();

	__try
	{
		sprintf_s(Buffer, "[%08x] [%c] [%s] ", GetTickCount(), Prefix, Subsystem ? Subsystem : "unknown");
		size_t BufLen = strlen(Buffer);

		vsnprintf_s(Buffer + BufLen, sizeof(Buffer) - BufLen, sizeof(Buffer) - BufLen, Text, Args);
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		CritSection.Leave();
		return;
	}

	__try
	{
		if (GetConsoleWindow())
			printf_s("%s\n", Buffer);
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
	}

	__try
	{
		if (LogFile)
		{
			fprintf_s(LogFile, "%s\n", Buffer);
			fflush(LogFile);
		}
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
	}

	CritSection.Leave();
}

#else

// Colored log messages
void System::Log::PrintMessage(const char * Subsystem, const char Prefix, const char * Text, va_list Args)
{
	CritSection.Enter();

	__try
	{
		if (GetConsoleWindow())
		{
			HANDLE Console = GetStdHandle(STD_OUTPUT_HANDLE);
			SetConsoleTextAttribute(Console, FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
			printf_s("[");
			SetConsoleTextAttribute(Console, FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE);
			printf_s("%08x", GetTickCount());
			SetConsoleTextAttribute(Console, FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
			printf_s("] [");

			WORD Attributes = FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE;
			switch(Prefix)
			{
			case 'M':
				Attributes = FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY;
				break;
			case 'W':
				Attributes = FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_INTENSITY;
				break;
			case 'E':
				Attributes = FOREGROUND_RED | FOREGROUND_INTENSITY;
				break;
			case 'C':
				Attributes = FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_INTENSITY;
				break;
			}
			SetConsoleTextAttribute(Console, Attributes);
			printf_s("%c", Prefix);
			SetConsoleTextAttribute(Console, FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
			printf_s("] [");
			SetConsoleTextAttribute(Console, Attributes);
			printf_s("%s", Subsystem ? Subsystem : "unknown");
			SetConsoleTextAttribute(Console, FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
			printf_s("] ");
			SetConsoleTextAttribute(Console, Attributes);
			vsnprintf_s(Buffer, sizeof(Buffer), sizeof(Buffer), Text, Args);
			printf_s("%s", Buffer);
			SetConsoleTextAttribute(Console, FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE);
			printf_s("\n");
		}
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
	}

	__try
	{
		if (LogFile)
		{
			sprintf_s(Buffer, "[%08x] [%c] [%s] ", GetTickCount(), Prefix, Subsystem ? Subsystem : "unknown");
			size_t BufLen = strlen(Buffer);

			vsnprintf_s(Buffer + BufLen, sizeof(Buffer) - BufLen, sizeof(Buffer) - BufLen, Text, Args);

			fprintf_s(LogFile, "%s\n", Buffer);
			fflush(LogFile);
		}
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
	}

	CritSection.Leave();
}

#endif

bool System::Log::AssignLogFile(const char * Filename)
{
	CloseLogFile();

	CritSection.Enter();

	if (fopen_s(&LogFile, Filename, "w+"))
	{
		LogFile = nullptr;
		CritSection.Leave();
		Error(Subsystem, "Cannot open file \"%s\" for log output", Filename);
		return false;
	}

	atexit(CloseLogFile);

	CritSection.Leave();

	return true;
}

void System::Log::CloseLogFile()
{
	CritSection.Enter();

	if(LogFile)
	{
		__try
		{
			fclose(LogFile);
			LogFile = nullptr;
		}
		__except(EXCEPTION_EXECUTE_HANDLER)
		{
		}
	}

	CritSection.Leave();
}

#ifdef CALL_PRINT_MESSAGE
#undef CALL_PRINT_MESSAGE
#endif
#define CALL_PRINT_MESSAGE(Subsystem, Prefix) \
	va_list Args; \
	va_start(Args, Text); \
	PrintMessage(Subsystem, Prefix, Text, Args); \
	va_end(Args);

void System::Log::Message(const char * Text, ...)
{
	if ((Level & MessageType::Message) == 0) return;

	CALL_PRINT_MESSAGE(nullptr, 'M')
}

void System::Log::Warning(const char * Text, ...)
{
	if ((Level & MessageType::Warning) == 0) return;

	CALL_PRINT_MESSAGE(nullptr, 'W')
}

void System::Log::Error(const char * Text, ...)
{
	if ((Level & MessageType::Error) == 0) return;

	CALL_PRINT_MESSAGE(nullptr, 'E')

	if (StopOnError) System::Application::Terminate();
}

void System::Log::CriticalError(const char * Text, ...)
{
	if ((Level & MessageType::CriticalError) == 0) return;

	CALL_PRINT_MESSAGE(nullptr, 'C')

	System::Application::Terminate();
}

void System::Log::Message(const char * Subsystem, const char * Text, ...)
{
	if ((Level & MessageType::Message) == 0) return;

	CALL_PRINT_MESSAGE(Subsystem, 'M')
}

void System::Log::Warning(const char * Subsystem, const char * Text, ...)
{
	if ((Level & MessageType::Warning) == 0) return;

	CALL_PRINT_MESSAGE(Subsystem, 'W')
}

void System::Log::Error(const char * Subsystem, const char * Text, ...)
{
	if ((Level & MessageType::Error) == 0) return;

	CALL_PRINT_MESSAGE(Subsystem, 'E')

	if (StopOnError) System::Application::Terminate();
}

void System::Log::CriticalError(const char * Subsystem, const char * Text, ...)
{
	if ((Level & MessageType::CriticalError) == 0) return;

	CALL_PRINT_MESSAGE(Subsystem, 'C')

	System::Application::Terminate();
}

void System::Log::Message(Types::Property<const char *> Subsystem, const char * Text, ...)
{
	if ((Level & MessageType::Message) == 0) return;

	CALL_PRINT_MESSAGE(Subsystem.Get(), 'M')
}

void System::Log::Warning(Types::Property<const char *> Subsystem, const char * Text, ...)
{
	if ((Level & MessageType::Warning) == 0) return;

	CALL_PRINT_MESSAGE(Subsystem.Get(), 'W')
}

void System::Log::Error(Types::Property<const char *> Subsystem, const char * Text, ...)
{
	if ((Level & MessageType::Error) == 0) return;

	CALL_PRINT_MESSAGE(Subsystem.Get(), 'E')

	if (StopOnError) System::Application::Terminate();
}

void System::Log::CriticalError(Types::Property<const char *> Subsystem, const char * Text, ...)
{
	if ((Level & MessageType::CriticalError) == 0) return;

	CALL_PRINT_MESSAGE(Subsystem.Get(), 'C')

	System::Application::Terminate();
}